import requests as www


class HTTPCacheInterface:
    __slots__ = ("base", "token", "ingress", "api", "name", "requests", "get_requests")

    def __init__(self, base: str, token: str, name=None):
        self.base = base
        self.token = token
        self.ingress = base + "/ingress"
        self.api = base + "/api"
        self.name = name or "http_cache_interface"
        self.requests = 0
        self.get_requests = 0

    def cached(self, url: str):
        return _cached(self, url)

    def cache(self, url: str):
        print(f"[{self.name}] Backend download for {url}")
        rq = www.get(self.api, params={"token": self.token, "url": url, "create": True})
        self.get_requests += 1
        rj = rq.json()
        return {"cached": rj["cached"], "url": rj["url"]}


class MDCacheInterface:
    __slots__ = ("base", "token", "ingress", "api", "name", "requests", "get_requests")

    def __init__(self, base: str, token: str, name=None):
        self.base = base
        self.token = token
        self.ingress = base + "/ingress"
        self.api = base + "/api"
        self.name = name or "md_cache_interface"
        self.requests = 0
        self.get_requests = 0

    def cached(self, url: str):
        return _cached(self, url)

    def cache(self, url: str):
        print(f"[{self.name}] Backend download for {url}")
        rq = www.get(self.ingress, params={"token": self.token, "url": url})
        self.get_requests += 1
        rj = rq.json()
        rj["md_url"] = rj.pop("url")
        cached_url = rj.pop("api_response")["url"]
        return {**rj, "url": cached_url}


def _cached(self, url):
    print(f"[{self.name}] Backend request for {url}")
    rq = www.get(self.api, params={"token": self.token, "url": url})
    self.requests += 1
    rj = rq.json()
    return {"cached": rj["cached"], "url": rj["url"]}



from time import time, sleep
from threading import Thread


def _time():
    return int(time())


class TimeScheduler:
    __slots__ = ("name", "checkInterval", "pendingOperations", "started", "thread", "auto")

    def __init__(self, name=None, interval: int = 1, pending: dict = None, auto: bool = True):
        self.name = name or "scheduler"
        self.checkInterval = interval
        self.pendingOperations = pending or {}
        self.started = False
        self.thread = None
        self.auto = auto

    def set(self, identifier, callback, args: list = None, kwargs: dict = None,
            time_left: int = None, time_set: int = None):
        if time_left:
            expected_time = _time() + time_left
        elif time_set:
            expected_time = time_set
        else:
            raise Exception("No schedule method selected. Use either time_left or time_set arguments.")
        self.pendingOperations[identifier] = (expected_time, callback, args, kwargs)
        print(f"[{self.name}] Added operation {identifier}")
        if self.auto:
            self._auto()

    def set_time(self, identifier, time_left: int = None, time_set: int = None):
        _op = self.pendingOperations[identifier]
        if time_left:
            expected_time = _time() + time_left
        elif time_set:
            expected_time = time_set
        else:
            raise Exception("No schedule method selected. Use either time_left or time_set arguments.")
        self.pendingOperations[identifier] = (expected_time, _op[1], _op[2], _op[3])
        print(f"[{self.name}] Set time for operation {identifier}")

    def set_time_relative(self, identifier, relative_time: int = None):
        _op = self.pendingOperations[identifier]
        self.pendingOperations[identifier] = (_op[0] + relative_time, _op[1], _op[2], _op[3])
        print(f"[{self.name}] Set time for operation {identifier}")

    def remove(self, identifier):
        if identifier in self.pendingOperations:
            print(f"[{self.name}] Removing operation {identifier}")
            return self.pendingOperations.pop(identifier)
        if self.auto:
            self._auto()

    def is_pending(self, identifier):
        if identifier in self.pendingOperations:
            return True
        else:
            return False

    def _target(self):
        while self.started is True:
            _to_remove = []
            _operations = self.pendingOperations.copy()
            for x in _operations:
                _op = _operations[x]
                if _op[0] <= _time():
                    print(f"[{self.name}] Processing {x}...")
                    _to_remove.append(x)
                    try:
                        if _op[2] is not None:
                            o = _op[1](*_op[2])
                        elif _op[3] is not None:
                            o = _op[1](**_op[3])
                        else:
                            o = _op[1]()
                    except Exception as e:
                        print(f"[{self.name}] {x} failed with error {e}")
                    else:
                        print(f"[{self.name}] {x} succeeded with output {o}")
            for y in _to_remove:
                self.pendingOperations.pop(y)
            if self.auto:
                self._auto()
            sleep(self.checkInterval)

    def start(self):
        self.started = True
        self.thread = Thread(target=self._target)
        self.thread.start()
        print(f"[{self.name}] Started scheduler thread")

    def stop(self):
        self.started = False
        print(f"[{self.name}] Stopped scheduler thread")

    def _auto(self):
        if self.pendingOperations and not self.started:
            self.start()
        elif not self.pendingOperations and self.started:
            self.stop()


class BucketScheduler:
    __slots__ = ("name", "bucketMidInterval", "pendingOperations", "defaultOperationTimer", "started", "thread", "auto")

    def __init__(self, name=None, interval: int = 0, timer: int = 1, pending: list = None, auto: bool = True):
        self.name = name or "scheduler"
        self.bucketMidInterval = interval
        self.defaultOperationTimer = timer
        self.pendingOperations = pending or []
        self.started = False
        self.thread = None
        self.auto = auto

    def add(self, callback, timer: int = None, args: list = None, kwargs: dict = None):
        timer = timer or self.defaultOperationTimer
        self.pendingOperations.append((timer, callback, args, kwargs))
        print(f"[{self.name}] Queued new operation")
        if self.auto:
            self._auto()

    def clear(self):
        self.pendingOperations = []
        print(f"[{self.name}] Cleared operation queue")
        if self.auto:
            self._auto()

    def _target(self):
        while self.started is True:
            if self.pendingOperations:
                _op = self.pendingOperations[0]
                print(f"[{self.name}] Processing operation...")
                try:
                    if _op[2] is not None:
                        o = _op[1](*_op[2])
                    elif _op[3] is not None:
                        o = _op[1](**_op[3])
                    else:
                        o = _op[1]()
                except Exception as e:
                    print(f"[{self.name}] Operation failed with error {e}")
                else:
                    print(f"[{self.name}] Operation succeeded with output {o}")
                self.pendingOperations.pop(0)
                sleep(_op[0])
            if self.auto:
                self._auto()
            sleep(self.bucketMidInterval)

    def start(self):
        self.started = True
        self.thread = Thread(target=self._target)
        self.thread.start()
        print(f"[{self.name}] Started bucket thread")

    def stop(self):
        self.started = False
        print(f"[{self.name}] Stopped bucket thread")

    def _auto(self):
        if self.pendingOperations and not self.started:
            self.start()
        elif not self.pendingOperations and self.started:
            self.stop()


class TimeRepeater:
    __slots__ = ("name", "checkInterval", "pendingOperations", "started", "thread", "auto")

    def __init__(self, name=None, interval: int = 1, pending: dict = None, auto: bool = True):
        self.name = name or "scheduler"
        self.checkInterval = interval
        self.pendingOperations = pending or {}
        self.started = False
        self.thread = None
        self.auto = auto

    def set(self, identifier, callback, interval: int, args: list = None, kwargs: dict = None):
        expected_time = _time() + interval
        self.pendingOperations[identifier] = (expected_time, interval, callback, args, kwargs)
        print(f"[{self.name}] Added operation {identifier}")
        if self.auto:
            self._auto()

    def remove(self, identifier):
        if identifier in self.pendingOperations:
            print(f"[{self.name}] Removing operation {identifier}")
            return self.pendingOperations.pop(identifier)
        if self.auto:
            self._auto()

    def is_pending(self, identifier):
        if identifier in self.pendingOperations:
            return True
        else:
            return False

    def _target(self):
        while self.started is True:
            for x in self.pendingOperations:
                _op = self.pendingOperations[x]
                if _op[0] <= _time():
                    print(f"[{self.name}] Processing {x}...")
                    try:
                        if _op[3] is not None:
                            o = _op[2](*_op[3])
                        elif _op[4] is not None:
                            o = _op[2](**_op[4])
                        else:
                            o = _op[2]()
                    except Exception as e:
                        print(f"[{self.name}] {x} failed with error {e}")
                    else:
                        print(f"[{self.name}] {x} succeeded with output {o}")
                    self.pendingOperations[x] = (_time() + _op[1], _op[1], _op[2], _op[3], _op[4])
            sleep(self.checkInterval)

    def start(self):
        self.started = True
        self.thread = Thread(target=self._target)
        self.thread.start()
        print(f"[{self.name}] Started scheduler thread")

    def stop(self):
        self.started = False
        print(f"[{self.name}] Stopped scheduler thread")

    def _auto(self):
        if self.pendingOperations and not self.started:
            self.start()
        elif not self.pendingOperations and self.started:
            self.stop()



import json
import requests


class H2SException(Exception):
    def __init__(self, r):
        self.status = r.status_code
        self.data = None
        try:
            self.data = r.json()
        except json.decoder.JSONDecodeError:
            self.data = r.content


class H2SNotFound(H2SException):
    pass


class H2SUnauthorized(H2SException):
    pass


class H2SAlreadyExists(H2SException):
    pass


class H2SClientError(H2SException):
    pass


class H2SServerError(H2SException):
    pass


class H2SInterface:
    __slots__ = ("server", "profile", "token", "base", "session")

    def __init__(self, server: str, profile: str, token: str):
        self.server = server
        self.profile = profile
        self.token = token
        self.base = server + "/" + profile
        self.session = requests.Session()
        self.session.headers["APIKey"] = token

    def reload_profiles(self):
        r = self.session.get(f"{self.server}/reload")
        if r.status_code == 200:
            return True
        else:
            self._handle_error_tree(r)

    def create_table(self, table: str):
        r = self.session.post(f"{self.base}/{table}")
        if r.status_code == 200:
            return True
        else:
            self._handle_error_tree(r)

    def delete_table(self, table: str):
        r = self.session.delete(f"{self.base}/{table}")
        if r.status_code == 200:
            return True
        else:
            self._handle_error_tree(r)

    def put(self, table: str, key: str, value: str, pool=None):
        h = {"Pool": pool} if pool else None
        r = self.session.put(f"{self.base}/{table}/{key}", headers=h, data=str(value).encode('utf-8'))
        if r.status_code == 200:
            return True
        else:
            self._handle_error_tree(r)

    def get(self, table: str, key: str, pool=None):
        h = {"Pool": pool} if pool else None
        r = self.session.get(f"{self.base}/{table}/{key}", headers=h)
        if r.status_code == 200:
            return r.content.decode('utf-8')
        elif r.status_code == 204:
            return None
        else:
            self._handle_error_tree(r)

    def delete(self, table: str, key: str, pool=None):
        h = {"Pool": pool} if pool else None
        r = self.session.delete(f"{self.base}/{table}/{key}", headers=h)
        if r.status_code == 200:
            return True
        else:
            self._handle_error_tree(r)

    def multi(self, data: dict):
        r = self.session.post(f"{self.base}/_multi", json=data)
        if r.status_code == 200:
            return r.json()
        elif r.status_code == 204:
            return None
        else:
            self._handle_error_tree(r)

    @staticmethod
    def _handle_error_tree(r):
        if r.status_code == 400:
            raise H2SClientError(r)
        elif r.status_code == 403:
            raise H2SUnauthorized(r)
        elif r.status_code == 404:
            raise H2SNotFound(r)
        elif r.status_code == 409:
            raise H2SAlreadyExists(r)
        elif r.status_code == 500:
            raise H2SServerError(r)
        else:
            raise H2SException(r)



import json
from pathlib import Path


class JSONDict(dict):
    def __init__(self, path, encoding='utf-8', data=None):
        self.path = Path(path)
        self.encoding = encoding
        if data:
            j = data
        elif self.path.is_file():
            with open(self.path, encoding=self.encoding, mode="r") as f:
                j = json.load(f)
        else:
            j = {}
        super().__init__(j)

    def save(self):
        with open(self.path, encoding=self.encoding, mode="w") as f:
            json.dump(self, f, indent=4, separators=(',', ': '))

    def reload(self):
        self.__init__(self.path, self.encoding)


class JSONList(list):
    def __init__(self, path, encoding='utf-8', data=None):
        self.path = Path(path)
        self.encoding = encoding
        if data:
            j = data
        elif self.path.is_file():
            with open(self.path, encoding=self.encoding, mode="r") as f:
                j = json.load(f)
        else:
            j = []
        super().__init__(j)

    def save(self):
        with open(self.path, encoding=self.encoding, mode="w") as f:
            json.dump(self, f, indent=4, separators=(',', ': '))

    def reload(self):
        self.__init__(self.path, self.encoding)



import json
import os
from random import randint


class DataIO:
    def save_json(self, filename, data):
        rnd = randint(1000, 9999)
        path, ext = os.path.splitext(filename)
        tmp_file = f"{path}-{rnd}.tmp"
        self._save_json(tmp_file, data)
        try:
            self._read_json(tmp_file)
        except json.decoder.JSONDecodeError:
            print(f"JSON integrity check on temp file for {filename} has failed. Original file unaltered.")
            return False
        os.replace(tmp_file, filename)
        return True

    def load_json(self, filename):
        return self._read_json(filename)

    def is_valid_json(self, filename):
        try:
            self._read_json(filename)
            return True
        except FileNotFoundError:
            return False
        except json.decoder.JSONDecodeError:
            return False

    @staticmethod
    def _read_json(filename):
        with open(filename, encoding='utf-8', mode="r") as f:
            data = json.load(f)
        return data

    @staticmethod
    def _save_json(filename, data):
        with open(filename, encoding='utf-8', mode="w") as f:
            json.dump(data, f, indent=4, separators=(',', ': '))
        return data

    @staticmethod
    def export_json(data):
        return json.dumps(data)

    @staticmethod
    def import_json(data):
        return json.loads(data)


dataIO = DataIO()



from time import time


def _time():
    return int(time())


class ObjectCache:
    __slots__ = ("data", "tracker", "lifespan", "name", "hits", "writes", "reads", "tests", "evictions")

    def __len__(self):
        return len(self.data)

    def __init__(self, data=None, tracker=None, lifespan=None, name=None):
        self.data = data or {}
        self.tracker = tracker or {}
        self.lifespan = lifespan or 3600
        self.name = name or "cache"
        self.hits = 0
        self.writes = 0
        self.reads = 0
        self.tests = 0
        self.evictions = 0

    def clear(self):
        print(f"[{self.name}] Clear cache memory.")
        self.__init__(name=self.name, lifespan=self.lifespan)

    def store(self, request, response):
        print(f"[{self.name}] Storing {request} at {_time()}")
        self.writes += 1
        self.data[request] = response
        self.tracker[request] = _time()
        return response

    def read(self, request):
        print(f"[{self.name}] Reading {request}")
        self.hits += 1
        self.reads += 1
        return self.data.get(request)

    def cached(self, request):
        self.tests += 1
        return request in self.data

    def evict(self, request, force: bool = False):
        if _time() > self.tracker.get(request, 0) + self.lifespan or force:
            self.evictions += 1
            self.data.pop(request)
            self.tracker.pop(request)
            print(f"[{self.name}] Evicting {request}")

    def evict_scan(self):
        for x in self.tracker.copy():
            if _time() > self.tracker.get(x, 0) + self.lifespan:
                self.evictions += 1
                self.data.pop(x)
                self.tracker.pop(x)
                print(f"[{self.name}] Evicting {x}")



class RoundRobin:
    def __init__(self, items, name=None):
        self.name = name or "rr"
        self.items = items
        self.item = 0

    def next(self):
        self.item = 0 if len(self.items) == self.item + 1 else self.item + 1
        print(f"[{self.name}] #{self.item}")
        return self.items[self.item]

    n = next



import img2pdf
import os
import sys
from pathlib import Path
valid_image_extensions = [".jpg", ".jpeg", ".png", ".gif", ".webp", ".tiff", ".bmp", ".heif", ".heic"]


def make_pdf(in_dir, out_file=None, force_conv=False):
    files = []
    cur_path = Path(in_dir)
    out_name = cur_path.name
    out_file = Path(out_file) if out_file else Path(str(cur_path) + f"/{out_name}.pdf")

    for x in os.listdir(str(cur_path)):
        cur_file = Path(str(cur_path) + "/" + x)
        if cur_file.suffix.lower() in valid_image_extensions:
            files.append(str(cur_file))

    files.sort()
    if force_conv:
        files = convert_images(files)
    try:
        with out_file.open("wb") as f:
            f.write(img2pdf.convert(files))
    except Exception as e:
        print(f"Could not convert to a PDF, attempting to convert images")
        print(f"Error: {e}")
        files = convert_images(files)
        with out_file.open("wb") as f:
            f.write(img2pdf.convert(files))
        [os.remove(y) for y in files if y.endswith("-pdf-conv")]
    return out_file


def convert_images(images):
    from PIL import Image
    n_files = []
    for i in images:
        fg = Image.open(i)
        if fg.mode in ("P", "RGBA", "LA", "PA"):
            fg.load()
            bg = Image.new("RGB", fg.size, (255, 255, 255))
            try:
                bg.paste(fg, mask=fg.split()[3])
            except IndexError:
                bg.paste(fg)
            bg.save(f"{i}-pdf-conv", "JPEG", quality=100)
            n_files.append(f"{i}-pdf-conv")
        else:
            n_files.append(i)
    return n_files


if __name__ == '__main__':
    try:
        gen = make_pdf(sys.argv[1])
        old_path = str(gen)
        new_path = str(Path(str(gen.parents[1]) + "/" + gen.name))
        os.rename(old_path, new_path)
    except IndexError:
        make_pdf(os.getcwd())



from .dataIO import dataIO
from .dataIO2 import JSONDict, JSONList
from .scheduler import BucketScheduler, TimeScheduler, TimeRepeater
from .object_cache import ObjectCache
from .php_cache import HTTPCacheInterface, MDCacheInterface
from .round_robin import RoundRobin
from .http2sql import H2SInterface

try:
    from .make_pdf import make_pdf
except ImportError:
    make_pdf = None
